package net.knarcraft.ffmpegconverter.converter;

import net.knarcraft.ffmpegconverter.container.FFMpegCommand;
import net.knarcraft.ffmpegconverter.container.ProcessResult;
import net.knarcraft.ffmpegconverter.container.StreamProbeResult;
import net.knarcraft.ffmpegconverter.converter.module.ConverterModule;
import net.knarcraft.ffmpegconverter.converter.module.DebugModule;
import net.knarcraft.ffmpegconverter.converter.module.ModuleExecutor;
import net.knarcraft.ffmpegconverter.converter.module.mapping.MapAllModule;
import net.knarcraft.ffmpegconverter.converter.module.output.CopyAudioModule;
import net.knarcraft.ffmpegconverter.converter.module.output.CopySubtitlesModule;
import net.knarcraft.ffmpegconverter.converter.module.output.SetOutputFileModule;
import net.knarcraft.ffmpegconverter.streams.AudioStream;
import net.knarcraft.ffmpegconverter.streams.StreamObject;
import net.knarcraft.ffmpegconverter.utility.FFMpegHelper;
import net.knarcraft.ffmpegconverter.utility.StringUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * A converter that crops letter-boxing (black padding) from video files
 */
public class LetterboxCropper extends AbstractConverter {

    private final static String SPACER = "|,-,|";

    /**
     * Initializes a new letter box cropper
     *
     * @param ffprobePath <p>Path/command to ffprobe</p>
     * @param ffmpegPath  <p>Path/command to ffmpeg</p>
     */
    public LetterboxCropper(@NotNull String ffprobePath, @NotNull String ffmpegPath) {
        super("mkv");
        this.ffprobePath = ffprobePath;
        this.ffmpegPath = ffmpegPath;
    }

    @Override
    public @NotNull List<String> getValidFormats() {
        return List.of("mkv", "mp4");
    }

    @Override
    @Nullable
    public FFMpegCommand generateConversionCommand(@NotNull String executable, @NotNull StreamProbeResult probeResult,
                                                   @NotNull String outFile) {
        File inputFile = probeResult.parsedFiles().get(0);

        List<ConverterModule> modules = new ArrayList<>();

        Map<String, Integer> cropValues = getLetterboxCropValues(modules, inputFile);

        FFMpegCommand convertCommand = getConversionCommand(cropValues, inputFile);

        modules.add(new MapAllModule<>(probeResult.getVideoStreams()));

        // Map audio if present
        List<AudioStream> audioStreams = probeResult.getAudioStreams();
        if (!audioStreams.isEmpty()) {
            modules.add(new MapAllModule<>(audioStreams));
            setOutputIndexes(audioStreams);
            modules.add(new CopyAudioModule(audioStreams));
        }

        // Map subtitles if present
        List<StreamObject> subtitleStreams = new ArrayList<>(probeResult.getSubtitleStreams());
        if (!subtitleStreams.isEmpty()) {
            modules.add(new MapAllModule<>(probeResult.getSubtitleStreams()));
            setOutputIndexes(subtitleStreams);
            modules.add(new CopySubtitlesModule());
        }

        // Map all attached streams, such as fonts and covers
        modules.add(new MapAllModule<>(probeResult.getOtherStreams()));

        modules.add(new SetOutputFileModule(outFile));

        new ModuleExecutor(convertCommand, modules).execute();
        return convertCommand;
    }

    /**
     * Gets the initial command used for removing the letterbox from a video file
     *
     * @param cropValues <p>The possible crop values to be used</p>
     * @param inputFile  <p>The file to be converted</p>
     * @return <p>The initial command to use for converting the file</p>
     */
    @NotNull
    private FFMpegCommand getConversionCommand(@NotNull Map<String, Integer> cropValues, @NotNull File inputFile) {
        String crop = null;
        int counts = 0;

        for (Map.Entry<String, Integer> entry : cropValues.entrySet()) {
            if (crop == null || entry.getValue() > counts) {
                crop = entry.getKey();
                counts = entry.getValue();
            }
        }

        if (crop == null) {
            throw new IllegalArgumentException("Unable to detect letterbox for video");
        }

        FFMpegCommand convertCommand = new FFMpegCommand(ffmpegPath);

        convertCommand.addInputFile(inputFile.getName());
        convertCommand.addOutputFileOption("-vf", "crop=" + crop);
        convertCommand.addOutputFileOption("-c:v", "libx265");
        convertCommand.addOutputFileOption("-crf", "22");
        convertCommand.addOutputFileOption("-preset", "medium");
        convertCommand.addOutputFileOption("-tune", "ssim");
        convertCommand.addOutputFileOption("-profile:v", "main10");
        convertCommand.addOutputFileOption("-level:v", "4");
        convertCommand.addOutputFileOption("-x265-params", "bframes=8:scenecut-bias=0.05:me=star:subme=5:" +
                "refine-mv=1:limit-refs=1:rskip=0:max-merge=5:rc-lookahead=80:lookahead-slices=5:min-keyint=23:" +
                "max-luma=1023:psy-rd=2:strong-intra-smoothing=0:b-intra=1:hist-threshold=0.03");
        return convertCommand;
    }

    /**
     * Gets a map counting each time a given letterbox crop suggestion has been encountered
     *
     * <p>As some parts of a video might be entirely black, or some parts of a video might be dark enough that parts of
     * a scene might be detected as part of the letterbox, this takes 30 crop-detect samples evenly spread among the
     * file. It can be assumed that the crop-detect value that was encountered the most times is the correct one.</p>
     *
     * @param modules   <p>The converter modules to append to</p>
     * @param inputFile <p>The input file to find letterbox of</p>
     * @return <p>A map counting all returned crop-detect values</p>
     */
    @NotNull
    private Map<String, Integer> getLetterboxCropValues(@NotNull List<ConverterModule> modules, @NotNull File inputFile) {
        Map<String, Integer> cropValues = new HashMap<>();

        if (this.debug) {
            modules.add(new DebugModule());
        }

        FFMpegCommand probeCommand = new FFMpegCommand(ffmpegPath);
        probeCommand.addGlobalOption("-nostats", "-hide_banner");
        probeCommand.addInputFile(inputFile.toString());
        probeCommand.addOutputFileOption("-vframes", "10");
        probeCommand.addOutputFileOption("-vf", "cropdetect");
        probeCommand.addOutputFileOption("-f", "null");
        probeCommand.addOutputFileOption("-");

        double duration;
        try {
            duration = FFMpegHelper.getDuration(ffprobePath, inputFile);
        } catch (IOException | NumberFormatException exception) {
            throw new RuntimeException("Unable to get duration from video file");
        }

        int increments = (int) (duration / 30d);

        for (int i = 0; i < duration; i += increments) {
            FFMpegCommand clone = probeCommand.clone();
            clone.addInputFileOption("-ss", String.valueOf(i));

            ProcessBuilder processBuilder = new ProcessBuilder(clone.getResult());
            ProcessResult result = null;
            try {
                result = FFMpegHelper.runProcess(processBuilder, inputFile.getParentFile(), SPACER, false);
            } catch (IOException ignored) {
            }
            if (result == null || result.exitCode() != 0) {
                throw new IllegalArgumentException("File probe failed with code " + (result == null ? "null" :
                        result.exitCode()));
            }

            List<String> parsed = StringUtil.stringBetween(result.output(), "crop=", SPACER);
            for (String string : parsed) {
                if (string.contains("-")) {
                    continue;
                }

                if (cropValues.containsKey(string)) {
                    cropValues.put(string, cropValues.get(string) + 1);
                } else {
                    cropValues.put(string, 1);
                }
            }
        }

        return cropValues;
    }

}
