package net.knarcraft.ffmpegconverter.converter.sorter;

import net.knarcraft.ffmpegconverter.streams.StreamObject;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.ArrayList;
import java.util.List;

/**
 * An abstract stream sorter, making implementation easier
 *
 * @param <L> <p>The type of streams this sorter sorts</p>
 */
public abstract class AbstractSorter<L extends StreamObject> implements StreamSorter<L> {

    protected StreamSorter<L> nextSorter = null;

    @Override
    public @NotNull StreamSorter<L> prepend(@NotNull StreamSorter<L> other) {
        this.nextSorter = other;
        return this;
    }

    @Override
    public @NotNull StreamSorter<L> append(@NotNull StreamSorter<L> other) {
        StreamSorter<L> end = other;
        while (end != null && end.hasChainElement()) {
            end = end.getNextInChain();
        }
        if (end == null) {
            throw new IllegalStateException("Other cannot be null. Something is wrong!");
        }
        end.setNextInChain(this);
        return other;
    }

    @Override
    public @NotNull List<L> chainSort(@NotNull List<L> input) {
        List<L> sorted = this.sort(input);
        if (nextSorter != null) {
            return nextSorter.chainSort(sorted);
        } else {
            return sorted;
        }
    }

    @Override
    public boolean hasChainElement() {
        return this.nextSorter != null;
    }

    @Override
    @Nullable
    public StreamSorter<L> getNextInChain() {
        return this.nextSorter;
    }

    @Override
    public void setNextInChain(@NotNull StreamSorter<L> next) {
        this.nextSorter = next;
    }

    /**
     * Sorts subtitle streams according to chosen languages and removes non-matching languages
     *
     * @param streams   <p>A list of streams to sort.</p>
     * @param languages <p>The languages chosen by the user.</p>
     * @param <G>       <p>The type of streams to sort.</p>
     * @return <p>A sorted version of the list.</p>
     */
    @NotNull
    protected <G extends StreamObject> List<G> sortStreamsByLanguage(@NotNull List<G> streams,
                                                                     @NotNull List<String> languages) {
        List<G> sorted = new ArrayList<>();
        for (String language : languages) {
            for (G stream : streams) {
                String streamLanguage = stream.getLanguage();
                if (language.equals("*") || (streamLanguage.equals("und") && language.equals("0")) ||
                        streamLanguage.equals(language)) {
                    sorted.add(stream);
                }
            }
            streams.removeAll(sorted);
        }
        return sorted;
    }

}
